from pathlib import Path
import sys
import itertools
import os
from os.path import isfile, join
import logging

from circleguard.loader import Loader
from circleguard.comparer import Comparer
from circleguard.investigator import Investigator
from circleguard.cacher import Cacher
from circleguard import config
from circleguard.exceptions import CircleguardException
from circleguard.replay import Check, ReplayMap, ReplayPath
from circleguard.enums import Detect, RatelimitWeight


class Circleguard:
    """
    Circleguard compares and investigates replays to detect cheats.

    Circleguard provides convenience methods for common use cases: map_check, verify, user_check, and local_check -
    see each method for further documentation. If these convenience methods are not flexible enough for you, you will
    have to instantiate your own Check object and call circleguard#run(check).

    Under the hood, convenience methods simply instantiate a Check object and call circleguard#run(check). The run method
    returns a generator containing Result objects, which contains the result of each comparison of the replays. See the
    Result class for further documentation.
    """

    # used to distinguish log output for cg instances
    NUM = 1

    def __init__(self, key, db_path, loader=None):
        """
        Initializes a Circleguard instance.

        Args:
            String key: An osu API key.
            [Path or String] db_path: A pathlike object to the databse file to write and/or read cached replays.
        """

        self.log = logging.getLogger(__name__ + str(Circleguard.NUM))
        self.db_path = Path(db_path).absolute() # allows for . to be passed to db_path
        cacher = Cacher(config.cache, self.db_path)
        # allow for people to pass their own loader implementation/subclass
        self.loader = Loader(cacher, key) if loader is None else loader(cacher, key)
        self.options = Options()
        Circleguard.NUM += 1

    def run(self, check):
        """
        Compares replays contained in the check object for replay steals.

        Args:
            Check check: A Check object containing either one or two sets of replays. If it was initialized with
                         a single replay set, all replays in that set are compared with each other. If it was
                         initialized with two replay sets, all replays in the first set are compared with all
                         replays in the second set.

        Returns:
            A generator containing Result objects of the comparisons.
        """

        self.log.info("Running circleguard with a Check")

        check.filter()
        # steal check
        compare1 = [replay for replay in check.replays if replay.detect & Detect.STEAL]
        compare2 = [replay for replay in check.replays2 if replay.detect & Detect.STEAL]
        num_to_load = len([replay for replay in compare1 + compare2 if replay.weight == RatelimitWeight.HEAVY])

        self.loader.new_session(num_to_load)
        check.load(self.loader)
        # all replays now have replay data, above is where ratelimit waiting would occur
        comparer = Comparer(check.thresh, compare1, replays2=compare2)
        yield from comparer.compare(mode=check.mode)

        # relax check (TODO)

    def map_check(self, map_id, u=None, num=None, cache=None, thresh=None, include=None):
        """
        Checks a map's leaderboard for replay steals.

        Args:
            Integer map_id: The id of the map (not the id of the mapset!) to compare replays from.
            Integer u: A user id. If passed, only the replay made by this user id on the given map will be
                       compared with the rest of the lederboard of the map. No other comparisons will be made.
            Integer num: The number of replays to compare from the map. Defaults to 50, or the config value if changed.
                         Loads from the top ranks of the leaderboard, so num=20 will compare the top 20 scores. This
                         number must be between 1 and 100, as restricted by the osu api.
            Boolean cache: Whether to cache the loaded replays. Defaults to False, or the config value if changed.
            Integer thresh: If a comparison scores below this value, its Result object has ischeat set to True.
                            Defaults to 18, or the config value if changed.
            Function include: A Predicate function that returns True if the replay should be loaded, and False otherwise.
                              The include function will be passed a single argument - the circleguard.Replay object, or one
                              of its subclasses.

        Returns:
            A generator containing Result objects of the comparisons.
        """
        check = self.create_map_check(map_id, u, num, cache, thresh, include)
        yield from self.run(check)


    def create_map_check(self, map_id, u=None, num=None, cache=None, thresh=None, include=None):
        """
        Creates the Check object used in the map_check convenience method. See that method for more information.
        """
        options = self.options
        num = num if num else options.num
        cache = cache if cache else options.cache
        thresh = thresh if thresh else options.thresh
        include = include if include else options.include

        self.log.info("Map check with map id %d, u %s, num %s, cache %s, thresh %s", map_id, u, num, cache, thresh)
        replays2 = None
        replay2_id = None
        if u:
            info = self.loader.user_info(map_id, user_id=u)
            replay2_id = info.replay_id
            replays2 = [ReplayMap(info.map_id, info.user_id, info.mods, username=info.username)]
        infos = self.loader.user_info(map_id, num=num)
        replays = []
        for info in infos:
            if info.replay_id == replay2_id:
                self.log.debug("Removing map %s, user %s, mods %s from map check with "
                                "the same replay id as the user's replay", info.map_id, info.user_id, info.mods)
                continue
            replays.append(ReplayMap(info.map_id, info.user_id, info.mods, username=info.username))
        return Check(replays, replays2=replays2, cache=cache, thresh=thresh, include=include)

    def verify(self, map_id, u1, u2, cache=None, thresh=None, include=None):
        """
        Verifies that two user's replay on a map are steals of each other.

        Args:
            Integer map_id: The id of the map to compare replays from.
            Integer u1: The user id of one of the users who set a replay on this map.
            Integer u2: The user id of the second user who set a replay on this map.
            Boolean cache: Whether to cache the loaded replays. Defaults to False, or the config value if changed.
            Integer thresh: If a comparison scores below this value, its Result object has ischeat set to True.
                            Defaults to 18, or the config value if changed.
            Function include: A Predicate function that returns True if the replay should be loaded, and False otherwise.
                              The include function will be passed a single argument - the circleguard.Replay object, or one
                              of its subclasses.

        Returns:
            A generator containing Result objects of the comparisons.
        """

        check = self.create_verify_check(map_id, u1, u2, cache, thresh, include)
        yield from self.run(check)

    def create_verify_check(self, map_id, u1, u2, cache=None, thresh=None, include=None):
        """
        Creates the Check object used in the verify_check convenience method. See that method for more information.
        """
        options = self.options
        cache = cache if cache else options.cache
        thresh = thresh if thresh else options.thresh
        include = include if include else options.include

        self.log.info("Verify with map id %d, u1 %s, u2 %s, cache %s", map_id, u1, u2, cache)
        info1 = self.loader.user_info(map_id, user_id=u1)
        info2 = self.loader.user_info(map_id, user_id=u2)
        replay1 = ReplayMap(info1.map_id, info1.user_id, info1.mods, username=info1.username)
        replay2 = ReplayMap(info2.map_id, info2.user_id, info2.mods, username=info2.username)

        return Check([replay1, replay2], cache=cache, thresh=thresh, include=include)

    def user_check(self, u, num, cache=None, thresh=None, include=None):
        """
        Checks a user's top plays for replay steals.

        For each of the user's top plays, the replay will be compared to the top plays of the map,
        then compared to all the user's other plays on the map, to check for both stealing and remodding.

        If a user has no or only one downloadable replay on the map, no comparisons to the user's other plays are made.
        Obviously, if the play is not downloadable, no comparison is made against the map leaderboard for that replay either.

        Args:
            Integer u: The user id of the user to check
            Integer num: The number of replays of each map to compare against the user's replay. For now, this also serves as the
                         number of top plays of the user to check for replay stealing and remodding.
            Boolean cache: Whether to cache the loaded replays. Defaults to False, or the config value if changed.
            Integer thresh: If a comparison scores below this value, its Result object has ischeat set to True.
                            Defaults to 18, or the config value if changed.
            Function include: A Predicate function that returns True if the replay should be loaded, and False otherwise.
                              The include function will be passed a single argument - the circleguard.Replay object, or one
                              of its subclasses.

        Returns:
            A generator containing Result objects of the comparisons.
        """

        for check_list in self.create_user_check(u, num, cache, thresh, include):
            # yuck; each top play has two different checks (remodding and stealing)
            # which is why we need a double loop
            for check in check_list:
                yield from self.run(check)

    def create_user_check(self, u, num_top, num_users, cache=None, thresh=None, include=None):
        """
        Creates the Check object used in the user_check convenience method. See that method for more information.

        Bewarned that this method does not return a single Check object like all other circleguard#create methods.
        Instead it returns a list of lists of Check objects [[Check, Check], [Check, Check], ...], because each top
        play of the user needs two Check objects, one for replay stealing and one for remodding. Be sure to handle
        this special case accordingly.
        """
        options = self.options
        cache = cache if cache else options.cache
        thresh = thresh if thresh else options.thresh
        include = include if include else options.include

        self.log.info("User check with u %s, num_top %s, num_users %s", u, num_top, num_users)
        ret = []
        for map_id in self.loader.get_user_best(u, num_top):
            info = self.loader.user_info(map_id, user_id=u)
            ureplay_id = info.replay_id # user replay id
            if not info.replay_available:
                continue  # if we can't download the user's replay on the map, we have nothing to compare against
            user_replay = [ReplayMap(info.map_id, info.user_id, mods=info.mods, username=info.username)]

            infos = self.loader.user_info(map_id, num=num_users)
            replays = []
            for info in infos:
                if info.replay_id == ureplay_id:
                    self.log.debug("Removing map %s, user %s, mods %s from user check with "
                                   "the same replay id as the user's replay", info.map_id, info.user_id, info.mods)
                    continue
                replays.append(ReplayMap(info.map_id, info.user_id, info.mods, username=info.username))

            remod_replays = []
            for info in self.loader.user_info(map_id, user_id=u, limit=False)[1:]:
                remod_replays.append(ReplayMap(info.map_id, info.user_id, mods=info.mods, username=info.username))

            check1 = Check(user_replay, replays2=replays, cache=cache, thresh=thresh, include=include)
            check2 = Check(user_replay + remod_replays, cache=cache, thresh=thresh, include=include)
            ret.append([check1, check2])

        return ret


    def local_check(self, folder, map_id=None, u=None, num=None, cache=None, thresh=None, include=None):
        """
        Compares locally stored osr files for replay steals.

        Args:
            [Path or String] folder: A pathlike object to the directory containing osr files.
            Integer map_id: A map id. If passed, the osr files will be compared against the top
                            plays on this map (50 by default).
            Integer u: A user id. If both this and map_id are passed, the osr files will be compared
                       against the user's play on the map. This value has no effect if passed without map_id.
            Integer num: The number of replays to compare from the map. Defaults to 50, or the config value if changed.
                         Loads from the top ranks of the leaderboard, so num=20 will compare the top 20 scores. This
                         number must be between 1 and 100, as restricted by the osu api. This value has no effect
                         if passed with both u and map_id.
            Boolean cache: Whether to cache the loaded replays. Defaults to False, or the config value if changed.
            Integer thresh: If a comparison scores below this value, its Result object has ischeat set to True.
                            Defaults to 18, or the config value if changed.
            Function include: A Predicate function that returns True if the replay should be loaded, and False otherwise.
                              The include function will be passed a single argument - the circleguard.Replay object, or one
                              of its subclasses.

        Returns:
            A generator containing Result objects of the comparisons.
        """

        check = self.create_local_check(folder, map_id, u, num, cache, thresh, include)
        yield from self.run(check)

    def create_local_check(self, folder, map_id=None, u=None, num=None, cache=None, thresh=None, include=None):
        """
        Creates the Check object used in the local_check convenience method. See that method for more information.
        """
        options = self.options
        num = num if num else options.num
        cache = cache if cache else options.cache
        thresh = thresh if thresh else options.thresh
        include = include if include else options.include

        paths = [folder / f for f in os.listdir(folder) if isfile(folder / f) and f.endswith(".osr")]
        local_replays = [ReplayPath(path) for path in paths]
        online_replays = None
        if map_id:
            if u:
                infos = self.loader.user_info(map_id, user_id=u)
            else:
                # num guaranteed to be defined, either passed or from settings.
                infos = self.loader.user_info(map_id, num=num)

            online_replays = [ReplayMap(info.map_id, info.user_id, info.mods, username=info.username) for info in infos]

        return Check(local_replays, replays2=online_replays, thresh=thresh, include=include)

    def load(self, check, replay):
        """
        Loads the given replay. This is identical to calling replay.load(cg.loader, check.cache) if cg is your
        Circleguard instance and check is your Check instance.. This method exists to emphasize that this behavior is encouraged,
        and tied to a specific cg (and Check) instance. The Check is necessary to inherit the cache setting from the Check
        in case it differs from the Circleguard option (since it is more specific, it would override circleguard). See the
        options documentation for more details on setting inheritence.
        """
        replay.load(self.loader, check.cache)

    def set_options(self, thresh=None, num=None, cache=None, failfast=None, loglevel=None, include=None):
        """
        Changes the default value for different options in circleguard.
        Affects only the ircleguard instance this method is called on.

        Args:
            Integer thresh: If a comparison scores below this value, its Result object has ischeat set to True. 18 by default.
            Integer num: How many replays to load from a map when doing a map check. 50 by default.
            Boolean cache: Whether downloaded replays should be cached or not. False by default.
            Boolean failfast: Will throw an exception if no comparisons can be made for a given Check object,
                          or silently make no comparisons otherwise. False by default.
            Integer loglevel: What level to log at. Circlecore follows standard python logging levels, with an added level of
                          TRACE with a value of 5 (lower than debug, which is 10). The value passed to loglevel is
                          passed directly to the setLevel function of this instance's logger. WARNING by default.
                          For more information on log levels, see the standard python logging lib.
            Function include: A Predicate function that returns True if the replay should be loaded, and False otherwise.
                          The include function will be passed a single argument - the circleguard.Replay object, or one
                          of its subclasses.
        """

        for k, v in locals().items():
            if v is None or k == "self":
                continue
            if k == "loglevel":
                self.log.setLevel(loglevel)
                continue
            if hasattr(self.options, k):
                setattr(self.options, k, v)
            else:  # this only happens if we fucked up, not the user's fault
                raise CircleguardException(f"The key {k} (value {v}) is not available as a config option for a circleguard instance")

def set_options(thresh=None, num=None, cache=None, failfast=None, loglevel=None, include=None):
    """
    Changes the default value for different options in circleguard.
    Affects all circleguard instances, even ones that have already been instantiated.

    Args:
        Integer thresh: If a comparison scores below this value, its Result object has ischeat set to True. 18 by default.
        Integer num: How many replays to load from a map when doing a map check. 50 by default.
        Boolean cache: Whether downloaded replays should be cached or not. False by default.
        Boolean failfast: Will throw an exception if no comparisons can be made for a given Check object,
                          or silently make no comparisons otherwise. False by default.
        Integer loglevel: What level to log at. Circlecore follows standard python logging levels, with an added level of
                          TRACE with a value of 5 (lower than debug, which is 10). The value passed to loglevel is
                          passed directly to the setLevel function of the circleguard root logger. WARNING by default.
                          For more information on log levels, see the standard python logging lib.
        Function include: A Predicate functrion that returns True if the replay should be loaded, and False otherwise.
                          The include function will be passed a single argument - the circleguard.Replay object, or one
                          of its subclasses.
    """

    for k, v in locals().items():
        if v is None:
            continue
        if k == "loglevel":
            logging.getLogger("circleguard").setLevel(loglevel)
            continue
        if hasattr(config, k):
            setattr(config, k, v)
        else:  # this only happens if we fucked up, not the user's fault
            raise CircleguardException(f"The key {k} (with value {v}) is not available as a config option for global config")



class Options():
    """
    Container class for options, tied to a specific Circleguard instance.
    """

    def __init__(self):
        ...

    # These methods are unfortunately necessary because when
    # config module variables are updated, references to them are not - ie
    # references to config.thresh (or any other) are by value. So, when we
    # access options attributes, just get the latest config variable with these
    # methods.
    @property
    def thresh(self):
        return config.thresh

    @property
    def num(self):
        return config.num

    @property
    def cache(self):
        return config.cache

    @property
    def failfast(self):
        return config.failfast

    @property
    def include(self):
        return config.include



import abc
import logging

import circleparse
import numpy as np

from circleguard.enums import Detect, RatelimitWeight
from circleguard import config
from circleguard.utils import TRACE

class Check():
    """
    Contains a list of Replay objects (or subclasses thereof) and how to proceed when
    investigating them for cheats.

    Attributes:
        List [Replay] replays: A list of Replay objects.
        List [Replay] replays2: A list of Replay objects to compare against 'replays' if passed.
        Integer thresh: If a comparison scores below this value, its Result object has ischeat set to True.
                        Defaults to 18, or the config value if changed.
        Boolean cache: Whether to cache the loaded replays. Defaults to False, or the config value if changed.
        String mode: "single" if only replays was passed, or "double" if both replays and replays2 were passed.
        Boolean loaded: False at instantiation, set to True once check#load is called. See check#load for
                        more details.
    """

    def __init__(self, replays, replays2=None, cache=None, thresh=None, include=None):
        """
        Initializes a Check instance.

        If only replays is passed, the replays in that list are compared with themselves. If
        both replays and replays2 are passed, the replays in replays are compared only with the
        replays in replays2. See comparer#compare for a more detailed description.

        Args:
            List [Replay] replays: A list of Replay objects.
            List [Replay] replays2: A list of Replay objects to compare against 'replays' if passed.
            Boolean cache: Whether to cache the loaded replays. Defaults to False, or the config value if changed.
            Integer thresh: If a comparison scores below this value, its Result object has ischeat set to True.
                            Defaults to 18, or the config value if changed.
        """

        self.log = logging.getLogger(__name__ + ".Check")
        self.replays = replays # list of ReplayMap and ReplayPath objects, not yet processed
        self.replays2 = replays2 if replays2 else [] # make replays2 fake iterable, for #filter mostly
        self.mode = "double" if replays2 else "single"
        self.loaded = False
        self.thresh = thresh if thresh else config.thresh
        self.cache = cache if cache else config.cache
        self.include = include if include else config.include

    def filter(self):
        """
        Filters self.replays and self.replays2 to contain only Replays where self.include returns True
        when the Replay is passed. This gives total control to what replays end up getting loaded
        and compared.
        """

        self.log.info("Filtering replays from Check")
        self.replays = [replay for replay in self.replays if self._include(replay)]
        self.replays2 = [replay for replay in self.replays2 if self._include(replay)]


    def _include(self, replay):
        """
        An internal helper method to create log statements from inside a list comprehension.
        """

        if(self.include(replay)):
            self.log.log(TRACE, "Replay passed include(), keeping in Check replays")
            return True
        else:
            self.log.debug("Replay failed include(), filtering from Check replays")
            return False

    def load(self, loader):
        """
        If check.loaded is already true, this method silently returns. Otherwise, loads replay data for every
        replay in both replays and replays2, and sets check.loaded to True. How replays are loaded is up to
        the implementation of the specific subclass of the Replay. Although the subclass may not use the loader
        object, it is still passed regardless to reduce type checking. For implementation details, see the load
        method of each Replay subclass.

        Args:
            Loader loader: The loader to handle api requests, if required by the Replay.
        """

        self.log.info("Loading replays from Check")

        if(self.loaded):
            self.log.debug("Check already loaded, not loading individual Replays")
            return
        for replay in self.replays:
            replay.load(loader, self.cache)
        if(self.replays2):
            for replay in self.replays2:
                replay.load(loader, self.cache)
        self.loaded = True
        self.log.debug("Finished loading Check object")

    def all_replays(self):
        """
        Convenience method for accessing all replays stored in this object.

        Returns:
            A list of all replays in this Check object (replays1 + replays2)
        """
        return self.replays + self.replays2


class Replay(abc.ABC):
    def __init__(self, username, mods, replay_id, replay_data, detect, weight):
        """
        Initializes a Replay instance.

        Args:
            String username: The username of the player who made the replay. Whether or not this is their true username
                             has no effect - this field is used to represent the player more readably than their id.
            Integer mods: The mods the replay was played with.
            Integer replay_id: The id of this replay, or 0 if it does not have an id (unsubmitted replays have no id)
            circleparse.Replay replay_data: A circleparse Replay containing the replay data for this replay. If the replay data is not available
                                         (from the api or otherwise), this field should be None. This means that this replay will not be
                                         compared against other replays or investigated for cheats.
            Detect detect: The Detect enum (or bitwise combination of enums), indicating what types of cheats this
                           replay should be investigated or compared for.
            RatelimitWeight weight: How much it 'costs' to load this replay from the api. If the load method of the replay makes no api calls,
                             this value is RatelimitWeight.NONE. If it makes only light api calls (anything but get_replay), this value is
                             RatelimitWeight.LIGHT. If it makes any heavy api calls (get_replay), this value is RatelimitWeight.HEAVY.
                             This value is used internally to determine how long the loader class will have to spend loading replays -
                             currently LIGHT and NONE are treated the same, and only HEAVY values are counted towards replays to load. Note
                             that this has no effect on the comparisons or internal program implementation - it only affects log messages
                             internally, and if you access circleguard#loader#total, it modifies that value as well. See Loader#new_session
                             for more details.
        """

        self.username = username
        self.mods = mods
        self.replay_id = replay_id
        self.replay_data = replay_data
        self.detect = detect
        self.weight = weight
        self.loaded = True


    @abc.abstractclassmethod
    def load(self, loader, cache):
        """
        Loads replay data of the replay, from the osu api or from some other source.
        Implementation is up to the specific subclass.

        To meet the specs of this method, subclasses must set replay.loaded to True after this method is called,
        and replay.replay_data must be a valid Replay object, as defined by circleparse.Replay. Both of these specs
        can be met if the superclass circleguard.Replay is initialized in the load method with a valid Replay, as
        circleguard.Replay.__init__ sets replay.loaded to true by default.
        """

        ...


    def as_list_with_timestamps(self):
        """
        Gets the playdata as a list of tuples of absolute time, x and y.

        Returns:
            A list of tuples of (t, x, y).
        """
        # get all offsets sum all offsets before it to get all absolute times
        timestamps = np.array([e.time_since_previous_action for e in self.replay_data])
        timestamps = timestamps.cumsum()

        # zip timestamps back to data and convert t, x, y to tuples
        txy = [[z[0], z[1].x, z[1].y] for z in zip(timestamps, self.replay_data)]
        # sort to ensure time goes forward as you move through the data
        # in case someone decides to make time go backwards anyway
        txy.sort(key=lambda p: p[0])
        return txy


class ReplayMap(Replay):
    """
    Represents a Replay submitted to online servers, that can be retrieved from the osu api.

    The only things you need to know to instantiate a ReplayMap
    are the user who made the replay, and the map it was made on.

    Attributes:
        Integer map_id: The id of the map the replay was made on.
        Integer user_id: The id of the user who made the replay.
        Integer mods: The mods the replay was played with. None if not set when instantiated and has not been loaded yet -
                      otherwise, set to the mods the replay was made with.
        String username: A readable representation of the user who made the replay. If passed,
                         username will be set to this string. Otherwise, it will be set to the user id.
                         This is so you don't need to know a user's username when creating a ReplayMap, only their id.
                         However, if the username is known (by retrieving it through the api, or other means), it is better
                         to represent the Replay with a player's name than an id. Both username and user_id will
                         obviously still be available to you through the result object after comparison.
        Detect detect: The Detect enum (or bitwise combination of enums), indicating what types of cheats this
                       replay should be investigated or compared for.
        Boolean loaded: Whether this replay has been loaded. If True, calls to #load will have no effect.
                        See #load for more information.
        RatelimitWeight weight: RatelimitWeight.HEAVY, as this class' load method makes a heavy api call. See RatelimitWeight
                                documentation for more information.
    """

    def __init__(self, map_id, user_id, mods=None, username=None, detect=Detect.ALL):
        """
        Initializes a ReplayMap instance.

        Args:
            Integer map_id: The id of the map the replay was made on.
            Integer user_id: The id of the user who made the replay.
            Integer mods: The mods the replay was played with. If this is not set, the top scoring replay of the user on the
                          given map will be loaded. Otherwise, the replay with the given mods will be loaded.
            String username: A readable representation of the user who made the replay. If passed,
                             username will be set to this string. Otherwise, it will be set to the user id.
                             This is so you don't need to know a user's username when creating a ReplayMap, only their id.
                             However, if the username is known (by retrieving it through the api, or other means), it is
                             better to represent the Replay with a player's name than an id. Both username and user_id
                             will obviously still be available to you through the result object after comparison.
            Detect detect: The Detect enum (or bitwise combination of enums), indicating what types of cheats this
                           replay should be investigated or compared for.
        """

        self.log = logging.getLogger(__name__ + ".ReplayMap")
        self.map_id = map_id
        self.user_id = user_id
        self.mods = mods
        self.detect = detect
        self.weight = RatelimitWeight.HEAVY
        self.loaded = False
        self.username = username if username else user_id

    def load(self, loader, cache=None):
        """
        Loads the data for this replay from the api. This method silently returns if replay.loaded is True.

        The superclass Replay is initialized after this call, setting replay.loaded to True. Multiple
        calls to this method will have no effect beyond the first.
        """
        self.log.debug("Loading ReplayMap for user %d on map %d with mods %d", self.user_id, self.map_id, self.mods)
        if(self.loaded):
            self.log.debug("ReplayMap already loaded, not loading")
            return
        info = loader.user_info(self.map_id, user_id=self.user_id, mods=self.mods)
        replay_data = loader.replay_data(info, cache=cache)
        Replay.__init__(self, self.username, info.mods, info.replay_id, replay_data, self.detect, self.weight)
        self.log.log(TRACE, "Finished loading ReplayMap")


class ReplayPath(Replay):
    """
    Represents a Replay saved locally.

    To instantiate a ReplayPath, you only need to know the path to the osr file. This class has significant
    advantages compared to a ReplayMap - the username is immediately available from the replay, instead of requiring
    an extra api call. The time it takes to load the replay is also significantly less – especially if you factor in
    ratelimits – because there is no need to make a request to the api to retrieve the replay data, only read an osr
    file.

    Of course, to reap those benefits, it requires having the replay already downloaded, which isn't always ideal.

    Attributes:
        [String or Path] path: A pathlike object representing the absolute path to the osr file.
        Detect detect: The Detect enum (or bitwise combination of enums), indicating what types of cheats this
                       replay should be investigated or compared for.
        Boolean loaded: Whether this replay has been loaded. If True, calls to #load will have no effect.
                        See #load for more information.
        RatelimitWeight weight: RatelimitWeight.NONE, as this class' load method makes no api calls. See RatelimitWeight
                                documentation for more information.
    """

    def __init__(self, path, detect=Detect.ALL):
        """
        Initializes a ReplayPath instance.

        Args:
            [String or Path] path: A pathlike object representing the absolute path to the osr file.
            Detect detect: The Detect enum (or bitwise combination of enums), indicating what types of cheats this
                           replay should be investigated or compared for.
        """

        self.log = logging.getLogger(__name__ + ".ReplayPath")
        self.path = path
        self.detect = detect
        self.weight = RatelimitWeight.HEAVY
        self.loaded = False

    def load(self, loader, cache=None):
        """
        Loads the data for this replay from the osr file given by the path. See circleparse.parse_replay_file for
        implementation details. This method has no effect if replay.loaded is True.

        The superclass Replay is initialized after this call, setting replay.loaded to True. Multiple
        calls to this method will have no effect beyond the first.

        The cache argument here currently has no effect, and is only added for homogeneity with ReplayMap#load.
        """

        self.log.debug("Loading ReplayPath with path %s", self.path)
        if(self.loaded):
            self.log.debug("ReplayPath already loaded, not loading")
            return
        # no, we don't need loader for ReplayPath, but to reduce type checking when calling we make the method signatures homogeneous
        loaded = circleparse.parse_replay_file(self.path)
        Replay.__init__(self, loaded.player_name, loaded.mod_combination, loaded.replay_id, loaded.play_data, self.detect, self.weight)
        self.log.log(TRACE, "Finished loading ReplayPath")



from enum import Enum, Flag

from circleguard.exceptions import UnknownAPIException, RatelimitException, InvalidKeyException, ReplayUnavailableException
# strings taken from osu api error responses
# [api response, exception class type, details to pass to an exception]
class Error(Enum):
    NO_REPLAY         = ["Replay not available.", ReplayUnavailableException, "Could not find any replay data. Skipping"]
    RATELIMITED       = ["Requesting too fast! Slow your operation, cap'n!", RatelimitException, "We were ratelimited. Waiting it out"]
    RETRIEVAL_FAILED  = ["Replay retrieval failed.", ReplayUnavailableException, "Replay retrieval failed. Skipping"]
    INVALID_KEY       = ["Please provide a valid API key.", InvalidKeyException, "Please provide a valid api key"]
    UNKNOWN           = ["Unknown error.", UnknownAPIException, "Unknown error when requesting a replay. Please report this "
                                                                "to the developers at https://github.com/circleguard/circlecore"]

class Mod(Enum):
    NoMod          = NM = 0
    NoFail         = NF = 1
    Easy           = EZ = 2
    NoVideo        = NV = 4
    Hidden         = HD = 8
    HardRock       = HR = 16
    SuddenDeath    = SD = 32
    DoubleTime     = DT = 64
    Relax          = RL = 128
    HalfTime       = HT = 256
    Nightcore      = NC = 512
    Flashlight     = FL = 1024
    Autoplay       = CN = 2048
    SpunOut        = SO = 4096
    Autopilot      = AP = 8192
    Perfect        = PF = 16384
    Key4           = K4 = 32768
    Key5           = K5 = 65536
    Key6           = K6 = 131072
    Key7           = K7 = 262144
    Key8           = K8 = 524288
    keyMod         = KM = 1015808
    FadeIn         = FI = 1048576
    Random         = RD = 2097152
    LastMod        = LM = 4194304
    TargetPractice = TP = 8388608
    Key9           = K9 = 16777216
    Coop           = CO = 33554432
    Key1           = K1 = 67108864
    Key3           = K3 = 134217728
    Key2           = K2 = 268435456

class Detect(Flag):
                   # (in binary)
    STEAL = 1 << 0 # 0001
    RELAX = 1 << 1 # 0010
    REMOD = 1 << 2 # 0100

    ALL = STEAL | RELAX | REMOD
    NONE = 0

class RatelimitWeight(Enum):
    """
    How much it 'costs' to load a replay from the api. If the load method of a replay makes no api calls,
    the corresponding value is RatelimitWeight.NONE. If it makes only light api calls (anything but get_replay),
    the corresponding value is RatelimitWeight.LIGHT. If it makes any heavy api calls (get_replay), the
    corresponding value is RatelimitWeight.HEAVY.

    This value is used internally to determine how long the loader class will have to spend loading replays -
    currently LIGHT and NONE are treated the same, and only HEAVY values are counted towards replays to load.
    See loader#new_session and the Replay documentation for more details.
    """

    NONE  = "none"
    LIGHT = "light"
    HEAVY = "heavy"



from datetime import datetime
import time
import base64
import sys
import logging
from math import ceil

from requests import RequestException
import circleparse
import ossapi

from circleguard.replay import ReplayMap
from circleguard.user_info import UserInfo
from circleguard.enums import Error
from circleguard.exceptions import (InvalidArgumentsException, APIException, CircleguardException,
                        RatelimitException, InvalidKeyException, ReplayUnavailableException, UnknownAPIException)
from circleguard.utils import TRACE

def request(function):
    """
    Decorator intended to appropriately handle all request and api related exceptions.

    Also checks if we can refresh the time at which we started our requests because
    it's been more than RATELIMIT_RESET since the first request of the cycle.

    If we've refreshed our ratelimits, sets start_time to be the current datetime.
    """

    def wrapper(*args, **kwargs):
        difference = datetime.now() - Loader.start_time
        if(difference.seconds > Loader.RATELIMIT_RESET):
            Loader.start_time = datetime.now()
        # catch them exceptions boy
        ret = None
        self = args[0]
        try:
            ret = function(*args, **kwargs)
        except RatelimitException:
            self._enforce_ratelimit()
            # wrap function with the decorator then call decorator
            ret = request(function)(*args, **kwargs)
        except InvalidKeyException as e:
            raise CircleguardException("The given key is invalid")
        except RequestException as e:
            self.log.warning("Request exception: {}. Likely a network issue; sleeping for 5 seconds then retrying".format(e))
            time.sleep(5)
            ret = request(function)(*args, **kwargs)
        except ReplayUnavailableException as e:
            self.log.warning("We expected a replay from the api, but it was unable to deliver it: {}".format(e))
            ret = None
        return ret
    return wrapper


def check_cache(function):
    """
    Decorator that checks if the replay by the given user_id on the given map_id is already
    cached. If so, returns a Replay instance from the cached data instead of requesting it
    from the api. Otherwise, it calls the function as normal.

    Note that self and user_info must be the first and second arguments to
    the function respectively.

    Returns:
        A Replay instance from the cached replay if it was cached,
        or the return value of the function if not.
    """

    def wrapper(*args, **kwargs):
        self = args[0]
        user_info = args[1]

        lzma = self.cacher.check_cache(user_info.map_id, user_info.user_id, user_info.mods)
        if(lzma):
            replay_data = circleparse.parse_replay(lzma, pure_lzma=True).play_data
            self.loaded += 1
            return replay_data
        else:
            return function(*args, **kwargs)
    return wrapper

class Loader():
    """
    Manages interactions with the osu api, using the ossapi wrapper.

    if the api ratelimits the key, we wait until we refresh our ratelimits and retry
    the request. Because the api does not provide the time until the next refresh (and we
    do not periodically retry the key), if the key is ratelimited outside of this class,
    the class may wait more time than necessary for the key to refresh.
    """

    RATELIMIT_RESET = 60 # time in seconds until the api refreshes our ratelimits
    start_time = datetime.min # when we started our requests cycle


    def __init__(self, cacher, key):
        """
        Initializes a Loader instance.
        """

        self.log = logging.getLogger(__name__)
        self.total = None
        self.loaded = 0
        self.api = ossapi.ossapi(key)
        self.cacher = cacher


    def new_session(self, total):
        """
        Resets the loaded replays to 0, and sets the total to the passed total.

        Intended to be called every time the loader is used for a different set of replay loadings -
        since a Loader instance is passed around to Comparer and Investigator, each with different
        amounts of replays to load, making new sessions is necessary to keep progress logs correct.
        """

        self.log.debug("Starting a new session with total %d", total)
        self.loaded = 0
        self.total = total


    @request
    def user_info(self, map_id, num=None, user_id=None, mods=None, limit=True):
        """
        Returns a list of UserInfo objects containing a user's
        (user_id, username, replay_id, enabled mods, replay available)
        on a given map.

        If limit and user_id is set, it will return a single UserInfo object, not a list.

        Args:
            Integer map_id: The map id to get the replay_id from.
            Integer user_id: The user id to get the replay_id from.
            Boolean limit: If set, will only return a user's top score (top response). Otherwise, will
                          return every response (every score they set on that map under different mods)
            Integer mods: The mods the replay info to retieve were played with.
        """

        # we have to define a new variable to hold locals - otherwise when we call it twice inside the dict comprehension,
        # it rebinds to the comp scope and takes on different locals which is real bad. I spent many-a-hour figuring this out,
        # and if anyone has a more elegant solution I'm all ears.
        locals_ = locals()
        self.log.log(TRACE, "Loading user info on map %d with options %s",
                            map_id, {k: locals_[k] for k in locals_ if k != 'self'})

        if(num and (num > 100 or num < 1)):
            raise InvalidArgumentsException("The number of top plays to fetch must be between 1 and 100 inclusive!")

        if(not bool(user_id) ^ bool(num)):
            raise InvalidArgumentsException("One of either num or user_id must be passed, but not both")

        response = self.api.get_scores({"m": "0", "b": map_id, "limit": num, "u": user_id, "mods": mods})
        Loader.check_response(response)
                                                                    # yes, it's necessary to cast the str response to int before bool - all strings are truthy.
        infos = [UserInfo(map_id, int(x["user_id"]), str(x["username"]), int(x["score_id"]), int(x["enabled_mods"]), bool(int(x["replay_available"]))) for x in response]

        return infos[0] if (limit and user_id) else infos # limit only applies if user_id was set


    @request
    def get_user_best(self, user_id, number):
        """
        Gets the top 100 best plays for the given user.

        Args:
            String user_id: The user id to get best plays of.
            Integer number: The number of top plays to retrieve. Must be between 1 and 100.

        Returns:
            A list of Integer map_ids for the given number of the user's top plays.

        Raises:
            InvalidArgumentsException if number is not between 1 and 100 inclusive.
        """

        self.log.log(TRACE, "Retrieving the best %d plays of user %d", number, user_id)
        if(number < 1 or number > 100):
            raise InvalidArgumentsException("The number of best user plays to fetch must be between 1 and 100 inclusive!")
        response = self.api.get_user_best({"m": "0", "u": user_id, "limit": number})

        Loader.check_response(response)

        return [int(x["beatmap_id"]) for x in response]


    @request
    def load_replay_data(self, map_id, user_id, mods=None):
        """
        Queries the api for replay data from the given user on the given map, with the given mods.

        Args:
            UserInfo user_info: The UserInfo representing this replay.
        Returns:
            The lzma bytes (b64 decoded response) returned by the api, or None if the replay was not available.

        Raises:
            CircleguardException if the loader instance has had a new session made yet.
            APIException if the api responds with an error we don't know.
        """

        self.log.log(TRACE, "Requesting replay data by user %d on map %d with mods %s", user_id, map_id, mods)
        if(self.total is None):
            raise CircleguardException("loader#new_session(total) must be called after instantiation, before any replay data is loaded.")

        response = self.api.get_replay({"m": "0", "b": map_id, "u": user_id, "mods": mods})

        Loader.check_response(response)
        self.loaded += 1

        return base64.b64decode(response["content"])

    @check_cache
    def replay_data(self, user_info, cache=None):
        """
        Loads the replay data specified by the user info.

        Args:
            UserInfo user_info: The UserInfo object representing this replay.

        Returns:
            The play_data field of an circleparse.Replay instance created from the replay data received from the api,
            or None if the replay was not available.

        Raises:
            UnknownAPIException if replay_available was 1, but we did not receive replay data from the api.
        """

        user_id = user_info.user_id
        map_id = user_info.map_id
        mods = user_info.mods
        if(not user_info.replay_available):
            self.log.debug("Replay data by user %d on map %d with mods %s not available", user_id, map_id, mods)
            return None

        lzma_bytes = self.load_replay_data(map_id, user_id, mods)
        if(lzma_bytes is None):
            raise UnknownAPIException("The api guaranteed there would be a replay available, but we did not receive any data. "
                                     "Please report this to the devs, who will open an issue on osu!api if necessary.")
        parsed_replay = circleparse.parse_replay(lzma_bytes, pure_lzma=True)
        replay_data = parsed_replay.play_data
        self.cacher.cache(lzma_bytes, user_info, should_cache=cache)
        return replay_data

    @staticmethod
    def check_response(response):
        """
        Checks the given api response for any kind of error or unexpected response.

        Args:
            String response: The api-returned response to check.

        Raises:
            An Error corresponding to the type of error if there was an error. The mappings
            for the api error message and its corresponding error are in circleguard.enums.
        """

        if("error" in response):
            for error in Error:
                if(response["error"] == error.value[0]):
                    raise error.value[1](error.value[2])
            else:
                raise Error.UNKNOWN.value[1](Error.UNKNOWN.value[2]) # pylint: disable=unsubscriptable-object
                # pylint is dumb because Error is an enum and this is totally legal

    def _enforce_ratelimit(self):
        """
        Enforces the api ratelimit by sleeping the thread until it's safe to make requests again.
        """

        difference = datetime.now() - Loader.start_time
        seconds_passed = difference.seconds
        if(seconds_passed > Loader.RATELIMIT_RESET):
            self.log.debug("More than a minute has passed since our last ratelimit, not sleeping")
            return

        # sleep the remainder of the reset cycle so we guarantee it's been that long since the first request
        sleep_seconds = Loader.RATELIMIT_RESET - seconds_passed

        self.log.info("Ratelimited, sleeping for %s seconds. %d of %d replays loaded. "
            "ETA ~ %d min", sleep_seconds, self.loaded, self.total, ceil((self.total-self.loaded)/10))
        time.sleep(sleep_seconds)



import logging

from circleguard.circleguard import Circleguard, set_options
from circleguard.replay import Check, Replay, ReplayMap, ReplayPath
from circleguard.enums import Detect
from circleguard.utils import TRACE, ColoredFormatter

__version__ = "2.0.2"

logging.addLevelName(TRACE, "TRACE")
formatter = ColoredFormatter("[%(threadName)s][%(name)s][%(levelname)s]  %(message)s  (%(filename)s:%(lineno)s)")
handler_stream = logging.StreamHandler()
handler_stream.setFormatter(formatter)
logging.getLogger("circleguard").addHandler(handler_stream)



from circleguard.replay import Replay
class Result():
    """
    Result objects are returned from circleguard#run, and contain the two Replay objects that
    were compared, among other statistics about the comparison. Replay data can be accessed
    through the replay1 and replay2 objects.

    Attributes:
        Replay replay1: The first replay that was compared against the second.
        Replay replay2: The second replay that was compared against the first.
        Integer similarity: How similar the two replays are. This number should not directly
                        be linked to any particular meaning, such as average distance between
                        the two cursors, but lower numbers will always mean the replays are
                        more direct steals of each other.
        Boolean ischeat: Whether similarity is less than the threshold set for this comparison (usually the
                     threshold set in the config, but not if it was overriden for the specific comparison)
        String later_name: A string representation of the player who made the replay that was set later
                       (meaning, if it was stolen, it was almost certainly stolen from the other replay in
                       the comparison). This will usually be the username of the player, but it takes the
                       value of whatever is stored in the replay's username field. See the Replay
                       documentation for more information on what values could be stored in this field.
    """

    def __init__(self, replay1: Replay, replay2: Replay, similarity: int, ischeat: bool, later_name: str):
        """
        Initializes a Result instance.

        This method is for internal use only. Unless you know what you're doing, you should only ever
        receive this object from calling circleguard method, and not create it yourself.

        Args:
            Replay replay1: The first replay that was compared against the second.
            Replay replay2: The second replay that was compared against the first.
            Integer similarity: How similar the two replays are. This number should not directly
                        be linked to any particular meaning, such as average distance between
                        the two cursors, but lower numbers will always mean the replays are
                        more direct steals of each other.
            Boolean ischeat: Whether similarity is less than the threshold set for this comparison
                        (usually the threshold set in the config, but not if it was overriden for
                        the specific comparison)
            String later_name: A string representation of the player who made the replay that was set later
                       (meaning, if it was stolen, it was almost certainly stolen from the other replay in
                       the comparison). This will usually be the username of the player, but it takes the
                       value of whatever is stored in the replay's username field. See the Replay
                       documentation for more information on what values could be stored in this field.
        """

        self.replay1: Replay = replay1
        self.replay2: Replay = replay2
        self.similarity: int = similarity
        self.ischeat: bool = ischeat
        self.later_name: str = later_name



from distutils.core import setup
from setuptools import find_packages
from circleguard.__init__ import __version__

with open("README.md", "r") as readme:
    long_description = readme.read()

setup(
    name="circleguard",
    version=__version__,
    description="A player made and maintained cheat detection tool for osu!. "
                "Provides support for detecting replay stealing and remodding "
                "from a profile, map, or set of osr files.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords = ["osu!, python, cheat-detection, replay-stealing, remodding"],
    author="Liam DeVoe",
    author_email="orionldevoe@gmail.com",
    url="https://github.com/circleguard/circlecore",
    download_url = "https://github.com/circleguard/circlecore/tarball/v" + __version__,
    license = "MIT",
    packages=find_packages(),
    install_requires=[
        "circleparse >= 5.0.1",
        "ossapi >= 1.2.0",
        "wtc >= 1.1.3",
        "numpy",
        "requests"
    ]
)



import itertools
import sys
import logging
import math

import numpy as np

from circleguard.replay import Replay
from circleguard.enums import Mod
from circleguard.exceptions import InvalidArgumentsException, CircleguardException
import circleguard.utils as utils
from circleguard.result import Result
import circleguard.config as config

class Comparer:
    """
    A class for managing a set of replay comparisons.

    Attributes:
        Integer threshold: If a comparison scores below this value, the Result object is assigned a ischeat value of True.
        List replays1: A list of Replay instances to compare against replays2 if passed, or against itself if not.
        List replays2: A list of Replay instances to be compared against.

    See Also:
        Investigator
    """

    def __init__(self, threshold, replays1, replays2=None):
        """
        Initializes a Comparer instance.

        Note that the order of the two replay lists has no effect; they are only numbered for consistency.
        Comparing 1 to 2 is the same as comparing 2 to 1.

        Args:
            Integer threshold: If a comparison scores below this value, the Result object is assigned a ischeat value of True.
            List replays1: A list of Replay instances to compare against replays2 if passed, or against itself if not.
            List replays2: A list of Replay instances to be compared against.
        """

        self.log = logging.getLogger(__name__)
        self.threshold = threshold

        # filter beatmaps we had no data for - see Loader.replay_data and OnlineReplay.from_map
        self.replays1 = [replay for replay in replays1 if replay.replay_data is not None]
        self.replays2 = [replay for replay in replays2 if replay.replay_data is not None] if replays2 else None

    def compare(self, mode):
        """
        If mode is "double", compares all replays in replays1 against all replays in replays2.
        If mode is "single", compares all replays in replays1 against all other replays in replays1 (len(replays1) choose 2 comparisons).
        In both cases, yields Result objects containing the result of each comparison.

        Args:
            String mode: One of either "double" or "single", determining how to choose which replays to compare.

        Returns:
            A generator containing Result objects of the comparisons.

        Raises:
            CircleguardException if no comparisons could be made from the given replays (if replays1
            or replays2 is empty) and config.failfast is True. Otherwise, even if the replay lists are empty,
            silently returns.
        """

        self.log.info("Comparing replays with mode: %s", mode)
        self.log.debug("replays1: %s", self.replays1)
        self.log.debug("replays2: %s", self.replays2)

        #TODO: a little bit hacky and I don't think works 100% correctly, if mode is double but replays2 is None
        if(not self.replays1 or self.replays2 == []):
            if(config.failfast):
                raise CircleguardException("No comparisons could be made from the given replays")
            else:
                return

        if(mode == "double"):
            iterator = itertools.product(self.replays1, self.replays2)
        elif (mode == "single"):
            iterator = itertools.combinations(self.replays1, 2)
        else:
            raise InvalidArgumentsException("'mode' must be one of 'double' or 'single'")

        for replay1, replay2 in iterator:
            yield self._result(replay1, replay2)


    def _result(self, replay1, replay2):
        """
        Compares two replays and returns the result of that comparison.

        Args:
            Replay replay1: The first replay to compare against the second
            Replay replay2: The second replay to compare against the first

        Returns:
            A Result object, containing the results of the comparison.
        """

        result = Comparer._compare_two_replays(replay1, replay2)
        mean = result[0]
        sigma = result[1]
        ischeat = False
        if(mean < self.threshold):
            ischeat = True

        # if they were both set locally, we may not get replay ids to compare (replay_id is 0 if local)
        later_name = None
        if(replay1.replay_id != 0 and replay2.replay_id != 0):
            later_name = replay1.username if(replay1.replay_id > replay2.replay_id) else replay2.username
        return Result(replay1, replay2, mean, ischeat, later_name)

    @staticmethod
    def _compare_two_replays(replay1, replay2):
        """
        Compares two Replays and return their average distance
        and standard deviation of distances.
        """

        # get all coordinates in numpy arrays so that they're arranged like:
        # [ x_1 x_2 ... x_n
        #   y_1 y_2 ... y_n ]
        # indexed by columns first.
        data1 = replay1.as_list_with_timestamps()
        data2 = replay2.as_list_with_timestamps()

        # interpolate
        (data1, data2) = utils.interpolate(data1, data2)

        # remove time from each tuple
        data1 = [d[1:] for d in data1]
        data2 = [d[1:] for d in data2]

        mods1 = [Mod(mod_val) for mod_val in utils.bits(replay1.mods)]
        mods2 = [Mod(mod_val) for mod_val in utils.bits(replay2.mods)]
        flip1 = Mod.HardRock in mods1
        flip2 = Mod.HardRock in mods2
        if(flip1 ^ flip2): # xor, if one has hr but not the other
            for d in data1:
                d[1] = 384 - d[1]

        (mu, sigma) = Comparer._compute_data_similarity(data1, data2)
        return (mu, sigma)

    @staticmethod
    def _compute_data_similarity(data1, data2):
        """
        Finds the similarity and standard deviation between two datasets.

        Args:
            List data1: A list of tuples containing the (x, y) coordinate of points
            List data2: A list of tuples containing the (x, y) coordinate of points

        Returns:
            A tuple containing (similarity value, standard deviation) between the two datasets
        """

        data1 = np.array(data1)
        data2 = np.array(data2)

        # switch if the second is longer, so that data1 is always the longest.
        if len(data2) > len(data1):
            (data1, data2) = (data2, data1)

        shortest = len(data2)

        distance = data1[:shortest] - data2
        # square all numbers and sum over the second axis (add row 2 to row 1),
        # finally take the square root of each number to get all distances.
        # [ x_1 x_2 ... x_n   => [ x_1 ** 2 ... x_n ** 2
        #   y_1 y_2 ... y_n ] =>   y_1 ** 2 ... y_n ** 2 ]
        # => [ x_1 ** 2 + y_1 ** 2 ... x_n ** 2 + y_n ** 2 ]
        # => [ d_1 ... d_2 ]
        distance = (distance ** 2).sum(axis=1) ** 0.5

        mu, sigma = distance.mean(), distance.std()

        return (mu, sigma)
